from trilogy.core.models import (
    Concept,
    Environment,
    MergeStatement,
)
from trilogy.core.processing.nodes import MergeNode, NodeJoin, History
from trilogy.core.processing.nodes.base_node import concept_list_to_grain, StrategyNode
from typing import List

from trilogy.core.enums import JoinType, PurposeLineage
from trilogy.constants import logger
from trilogy.core.processing.utility import padding
from trilogy.core.processing.node_generators.common import concept_to_relevant_joins
from itertools import combinations
from trilogy.core.processing.node_generators.common import resolve_join_order
from trilogy.utility import unique


LOGGER_PREFIX = "[GEN_CONCEPT_MERGE_NODE]"


def merge_joins(
    parents: List[StrategyNode], merge_concepts: List[Concept]
) -> List[NodeJoin]:
    output = []
    for left, right in combinations(parents, 2):
        output.append(
            NodeJoin(
                left_node=left,
                right_node=right,
                concepts=merge_concepts,
                join_type=JoinType.FULL,
                filter_to_mutual=True,
            )
        )
    return resolve_join_order(output)


def gen_concept_merge_node(
    concept: Concept,
    local_optional: List[Concept],
    environment: Environment,
    g,
    depth: int,
    source_concepts,
    history: History | None = None,
) -> MergeNode | None:
    if not isinstance(concept.lineage, MergeStatement):
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} Cannot generate merge node for {concept}"
        )
        return None
    lineage: MergeStatement = concept.lineage

    base_parents: List[StrategyNode] = []

    # get additional concepts that should be merged across the environments
    additional_merge: List[Concept] = [*lineage.concepts]
    target_namespaces = set(x.namespace for x in [concept] + local_optional)
    for x in local_optional:
        if x.address in environment.merged_concepts:
            ms = environment.merged_concepts[x.address].lineage
            assert isinstance(ms, MergeStatement)
            additional_merge += ms.concepts

    for select in lineage.concepts:
        # if it's a merge concept, filter it out of the optional
        if select.namespace not in target_namespaces:
            continue
        sub_optional = [
            x
            for x in local_optional
            if x.address not in environment.merged_concepts
            and x.namespace == select.namespace
        ]

        sub_additional_merge = [
            x for x in additional_merge if x.namespace == select.namespace
        ]
        sub_optional += sub_additional_merge
        final: List[Concept] = unique([select] + sub_optional, "address")
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} generating concept merge parent node with {[x.address for x in final]}"
        )
        snode: StrategyNode = source_concepts(
            mandatory_list=final,
            environment=environment,
            g=g,
            depth=depth + 1,
            history=history,
        )

        if not snode:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} Cannot generate merge node for {concept}"
            )
            return None
        for x in sub_additional_merge:
            snode.add_output_concept(environment.merged_concepts[x.address])
        base_parents.append(snode)

    node_joins = merge_joins(
        base_parents,
        unique(
            [environment.merged_concepts[x.address] for x in additional_merge],
            "address",
        ),
    )

    enrichment = set([x.address for x in local_optional])
    outputs = [
        x
        for y in base_parents
        for x in y.output_concepts
        if x.derivation != PurposeLineage.MERGE
    ]

    additional_relevant = [x for x in outputs if x.address in enrichment]
    final_outputs = outputs + additional_relevant + [concept]
    virtual_outputs = [x for x in final_outputs if x.derivation == PurposeLineage.MERGE]
    node = MergeNode(
        input_concepts=[x for y in base_parents for x in y.output_concepts],
        output_concepts=[
            x for x in final_outputs if x.derivation != PurposeLineage.MERGE
        ],
        environment=environment,
        g=g,
        depth=depth,
        parents=base_parents,
        node_joins=node_joins,
        virtual_output_concepts=virtual_outputs,
    )

    qds = node.rebuild_cache()

    # assume grain to be outoput of select
    # but don't include anything aggregate at this point
    qds.grain = concept_list_to_grain(
        node.output_concepts, parent_sources=qds.datasources
    )
    possible_joins = concept_to_relevant_joins(additional_relevant)
    if not local_optional:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} no enriched required for merge concept node; exiting early"
        )
        return node
    if not possible_joins:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} no possible joins for merge concept node; exiting early"
        )
        return node
    if all(
        [x.address in [y.address for y in node.output_concepts] for x in local_optional]
    ):
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} all enriched concepts returned from base merge concept node; exiting early"
        )
        return node
    missing = [
        x
        for x in local_optional
        if x.address not in [y.address for y in node.output_concepts]
    ]
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} generating merge concept enrichment node for missing {[x.address for x in missing]}"
    )
    enrich_node: MergeNode = source_concepts(  # this fetches the parent + join keys
        # to then connect to the rest of the query
        mandatory_list=additional_relevant + missing,
        environment=environment,
        g=g,
        depth=depth + 1,
        history=history,
    )
    if not enrich_node:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} Cannot generate merge concept enrichment node for {concept.address} with optional {[x.address for x in local_optional]}, returning just merge concept"
        )
        return node

    # we still need the hidden concepts to be returned to the search
    # since they must be on the final node
    # to avoid further recursion
    # TODO: let the downstream search know they were found
    return MergeNode(
        input_concepts=enrich_node.output_concepts + node.output_concepts,
        # also filter out the
        output_concepts=[
            x
            for x in node.output_concepts + local_optional
            if x.derivation != PurposeLineage.MERGE
        ],
        hidden_concepts=[],
        environment=environment,
        g=g,
        depth=depth,
        parents=[
            # this node gets the window
            node,
            # this node gets enrichment
            enrich_node,
        ],
        node_joins=[
            NodeJoin(
                left_node=enrich_node,
                right_node=node,
                concepts=possible_joins,
                filter_to_mutual=False,
                join_type=JoinType.LEFT_OUTER,
            )
        ],
        partial_concepts=node.partial_concepts,
        virtual_output_concepts=virtual_outputs,
    )



from trilogy.core.models import (
    CTE,
    SelectStatement,
    PersistStatement,
    Datasource,
    MultiSelectStatement,
    Conditional,
    BooleanOperator,
)
from trilogy.core.enums import PurposeLineage
from trilogy.constants import logger, CONFIG
from abc import ABC


class OptimizationRule(ABC):

    def optimize(self, cte: CTE, inverse_map: dict[str, list[CTE]]) -> bool:
        raise NotImplementedError

    def log(self, message: str):
        logger.info(f"[Optimization][{self.__class__.__name__}] {message}")

    def debug(self, message: str):
        logger.debug(f"[Optimization][{self.__class__.__name__}] {message}")


class InlineDatasource(OptimizationRule):

    def optimize(self, cte: CTE, inverse_map: dict[str, list[CTE]]) -> bool:
        if not cte.parent_ctes:
            return False

        optimized = False
        self.log(
            f"Checking {cte.name} for consolidating inline tables with {len(cte.parent_ctes)} parents"
        )
        to_inline: list[CTE] = []
        force_group = False
        for parent_cte in cte.parent_ctes:
            if not parent_cte.is_root_datasource:
                self.log(f"parent {parent_cte.name} is not root")
                continue
            if parent_cte.parent_ctes:
                self.log(f"parent {parent_cte.name} has parents")
                continue
            raw_root = parent_cte.source.datasources[0]
            if not isinstance(raw_root, Datasource):
                self.log(f"parent {parent_cte.name} is not datasource")
                continue
            root: Datasource = raw_root
            if not root.can_be_inlined:
                self.log(f"parent {parent_cte.name} datasource is not inlineable")
                continue
            root_outputs = {x.address for x in root.output_concepts}
            cte_outputs = {x.address for x in parent_cte.output_columns}
            grain_components = {x.address for x in root.grain.components}
            if not cte_outputs.issubset(root_outputs):
                self.log(f"Not all {parent_cte.name} outputs are found on datasource")
                continue
            if not grain_components.issubset(cte_outputs):
                self.log("Not all datasource components in cte outputs, forcing group")
                force_group = True
            to_inline.append(parent_cte)

        for replaceable in to_inline:

            result = cte.inline_parent_datasource(replaceable, force_group=force_group)
            if result:
                self.log(f"Inlined parent {replaceable.name}")
            else:
                self.log(f"Failed to inline {replaceable.name}")
        return optimized


def decompose_condition(conditional: Conditional):
    chunks = []
    if conditional.operator == BooleanOperator.AND:
        for val in [conditional.left, conditional.right]:
            if isinstance(val, Conditional):
                chunks.extend(decompose_condition(val))
            else:
                chunks.append(val)
    else:
        chunks.append(conditional)
    return chunks


def is_child_of(a, comparison):
    if isinstance(comparison, Conditional):
        return (
            is_child_of(a, comparison.left) or is_child_of(a, comparison.right)
        ) and comparison.operator == BooleanOperator.AND
    return comparison == a


class PredicatePushdown(OptimizationRule):

    def optimize(self, cte: CTE, inverse_map: dict[str, list[CTE]]) -> bool:

        if not cte.parent_ctes:
            self.debug(f"No parent CTEs for {cte.name}")

            return False

        optimized = False
        if not cte.condition:
            self.debug(f"No CTE condition for {cte.name}")
            return False
        self.log(
            f"Checking {cte.name} for predicate pushdown with {len(cte.parent_ctes)} parents"
        )
        if isinstance(cte.condition, Conditional):
            candidates = cte.condition.decompose()
        else:
            candidates = [cte.condition]
        logger.info(f"Have {len(candidates)} candidates to try to push down")
        for candidate in candidates:
            conditions = {x.address for x in candidate.concept_arguments}
            for parent_cte in cte.parent_ctes:
                materialized = {k for k, v in parent_cte.source_map.items() if v != []}
                if conditions.issubset(materialized):
                    if all(
                        [
                            is_child_of(candidate, child.condition)
                            for child in inverse_map[parent_cte.name]
                        ]
                    ):
                        self.log(
                            f"All concepts are found on {parent_cte.name} and all it's children include same filter; pushing up filter"
                        )
                        if parent_cte.condition:
                            parent_cte.condition = Conditional(
                                left=parent_cte.condition,
                                operator=BooleanOperator.AND,
                                right=candidate,
                            )
                        else:
                            parent_cte.condition = candidate
                        optimized = True
                else:
                    logger.info("conditions not subset of parent materialized")

        if all(
            [
                is_child_of(cte.condition, parent_cte.condition)
                for parent_cte in cte.parent_ctes
            ]
        ):
            self.log("All parents have same filter, removing filter")
            cte.condition = None
            optimized = True

        return optimized


def filter_irrelevant_ctes(
    input: list[CTE],
    root_cte: CTE,
):
    relevant_ctes = set()

    def recurse(cte: CTE):
        relevant_ctes.add(cte.name)
        for cte in cte.parent_ctes:
            recurse(cte)

    recurse(root_cte)
    return [cte for cte in input if cte.name in relevant_ctes]


def gen_inverse_map(input: list[CTE]) -> dict[str, list[CTE]]:
    inverse_map: dict[str, list[CTE]] = {}
    for cte in input:
        for parent in cte.parent_ctes:
            if parent.name not in inverse_map:
                inverse_map[parent.name] = []
            inverse_map[parent.name].append(cte)
    return inverse_map


def is_direct_return_eligible(
    cte: CTE, select: SelectStatement | PersistStatement | MultiSelectStatement
) -> bool:
    if isinstance(select, (PersistStatement, MultiSelectStatement)):
        return False
    derived_concepts = [
        c for c in cte.source.output_concepts if c not in cte.source.input_concepts
    ]
    eligible = True
    conditions = (
        set(x.address for x in select.where_clause.concept_arguments)
        if select.where_clause
        else set()
    )
    if conditions and select.limit:
        return False
    for x in derived_concepts:
        if x.derivation == PurposeLineage.WINDOW:
            return False
        if x.derivation == PurposeLineage.UNNEST:
            return False
        if x.derivation == PurposeLineage.AGGREGATE:
            if x.address in conditions:
                return False
    logger.info(
        f"Upleveling output select to final CTE with derived_concepts {[x.address for x in derived_concepts]}"
    )
    return eligible


def sort_select_output(cte: CTE, query: SelectStatement | MultiSelectStatement):
    hidden_addresses = [c.address for c in query.hidden_components]
    output_addresses = [
        c.address for c in query.output_components if c.address not in hidden_addresses
    ]

    mapping = {x.address: x for x in cte.output_columns}

    new_output = []
    for x in output_addresses:
        new_output.append(mapping[x])
    cte.output_columns = new_output


def optimize_ctes(
    input: list[CTE], root_cte: CTE, select: SelectStatement | MultiSelectStatement
):
    complete = False
    REGISTERED_RULES: list["OptimizationRule"] = []

    if CONFIG.optimizations.datasource_inlining:
        REGISTERED_RULES.append(InlineDatasource())
    if CONFIG.optimizations.predicate_pushdown:
        REGISTERED_RULES.append(PredicatePushdown())

    while not complete:
        actions_taken = False
        for rule in REGISTERED_RULES:
            for cte in input:
                inverse_map = gen_inverse_map(input)
                actions_taken = rule.optimize(cte, inverse_map)
        complete = not actions_taken

    if CONFIG.optimizations.direct_return and is_direct_return_eligible(
        root_cte, select
    ):
        root_cte.order_by = select.order_by
        root_cte.limit = select.limit
        if select.where_clause:

            if root_cte.condition:
                root_cte.condition = Conditional(
                    left=root_cte.condition,
                    operator=BooleanOperator.AND,
                    right=select.where_clause.conditional,
                )
            else:
                root_cte.condition = select.where_clause.conditional
        root_cte.requires_nesting = False
        sort_select_output(root_cte, select)

    return filter_irrelevant_ctes(input, root_cte)



from trilogy.core.models import (
    Concept,
    Environment,
    SelectStatement,
    RowsetDerivationStatement,
    RowsetItem,
    MultiSelectStatement,
)
from trilogy.core.processing.nodes import MergeNode, NodeJoin, History, StrategyNode
from trilogy.core.processing.nodes.base_node import concept_list_to_grain
from typing import List

from trilogy.core.enums import JoinType, PurposeLineage
from trilogy.constants import logger
from trilogy.core.processing.utility import padding, unique
from trilogy.core.processing.node_generators.common import concept_to_relevant_joins


LOGGER_PREFIX = "[GEN_ROWSET_NODE]"


def gen_rowset_node(
    concept: Concept,
    local_optional: List[Concept],
    environment: Environment,
    g,
    depth: int,
    source_concepts,
    history: History | None = None,
) -> StrategyNode | None:
    if not isinstance(concept.lineage, RowsetItem):
        raise SyntaxError(
            f"Invalid lineage passed into rowset fetch, got {type(concept.lineage)}, expected {RowsetItem}"
        )
    lineage: RowsetItem = concept.lineage
    rowset: RowsetDerivationStatement = lineage.rowset
    select: SelectStatement | MultiSelectStatement = lineage.rowset.select
    parents: List[StrategyNode] = []
    if where := select.where_clause:
        targets = select.output_components + where.conditional.row_arguments
        for sub_select in where.conditional.existence_arguments:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} generating parent existence node with {[x.address for x in sub_select]}"
            )
            parent_check = source_concepts(
                mandatory_list=sub_select,
                environment=environment,
                g=g,
                depth=depth + 1,
                history=history,
            )
            if not parent_check:
                logger.info(
                    f"{padding(depth)}{LOGGER_PREFIX} Cannot generate parent existence node for rowset node for {concept}"
                )
                return None
            parents.append(parent_check)
    else:
        targets = select.output_components
    node: StrategyNode = source_concepts(
        mandatory_list=unique(targets, "address"),
        environment=environment,
        g=g,
        depth=depth + 1,
        history=history,
    )

    # add our existence concepts in
    if parents:
        node.parents += parents
        for parent in parents:
            for x in parent.output_concepts:
                if x.address not in node.output_lcl:
                    node.existence_concepts.append(x)
    if not node:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} Cannot generate rowset node for {concept}"
        )
        return None
    node.conditions = select.where_clause.conditional if select.where_clause else None
    enrichment = set([x.address for x in local_optional])
    rowset_relevant = [x for x in rowset.derived_concepts]
    select_hidden = set([x.address for x in select.hidden_components])
    rowset_hidden = [
        x
        for x in rowset.derived_concepts
        if isinstance(x.lineage, RowsetItem)
        and x.lineage.content.address in select_hidden
    ]
    additional_relevant = [
        x for x in select.output_components if x.address in enrichment
    ]
    # add in other other concepts
    for item in rowset_relevant:
        node.output_concepts.append(item)
    for item in additional_relevant:
        node.output_concepts.append(item)
    if select.where_clause:
        for item in additional_relevant:
            node.partial_concepts.append(item)
    node.hidden_concepts = rowset_hidden + [
        x
        for x in node.output_concepts
        if x.address not in [y.address for y in local_optional + [concept]]
        and x.derivation != PurposeLineage.ROWSET
    ]
    # assume grain to be output of select
    # but don't include anything aggregate at this point
    node.rebuild_cache()
    assert node.resolution_cache

    node.resolution_cache.grain = concept_list_to_grain(
        node.output_concepts, parent_sources=node.resolution_cache.datasources
    )

    possible_joins = concept_to_relevant_joins(additional_relevant)
    if not local_optional:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} no enriched required for rowset node; exiting early"
        )
        return node
    if not possible_joins:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} no possible joins for rowset node; exiting early"
        )
        return node
    if all(
        [x.address in [y.address for y in node.output_concepts] for x in local_optional]
    ):
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} all enriched concepts returned from base rowset node; exiting early"
        )
        return node
    enrich_node: MergeNode = source_concepts(  # this fetches the parent + join keys
        # to then connect to the rest of the query
        mandatory_list=additional_relevant + local_optional,
        environment=environment,
        g=g,
        depth=depth + 1,
    )
    if not enrich_node:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} Cannot generate rowset enrichment node for {concept} with optional {local_optional}, returning just rowset node"
        )
        return node

    return MergeNode(
        input_concepts=enrich_node.output_concepts + node.output_concepts,
        output_concepts=node.output_concepts + local_optional,
        environment=environment,
        g=g,
        depth=depth,
        parents=[
            # this node gets the window
            node,
            # this node gets enrichment
            enrich_node,
        ],
        node_joins=[
            NodeJoin(
                left_node=enrich_node,
                right_node=node,
                concepts=concept_to_relevant_joins(additional_relevant),
                filter_to_mutual=False,
                join_type=JoinType.LEFT_OUTER,
            )
        ],
        partial_concepts=node.partial_concepts,
    )



from enum import Enum

InfiniteFunctionArgs = -1


class UnnestMode(Enum):
    DIRECT = "direct"
    CROSS_APPLY = "cross_apply"
    CROSS_JOIN = "cross_join"


class ConceptSource(Enum):
    MANUAL = "manual"
    AUTO_DERIVED = "auto_derived"


class StatementType(Enum):
    QUERY = "query"


class Purpose(Enum):
    CONSTANT = "const"
    KEY = "key"
    PROPERTY = "property"
    METRIC = "metric"
    ROWSET = "rowset"
    AUTO = "auto"

    @classmethod
    def _missing_(cls, value):
        if value == "constant":
            return Purpose.CONSTANT
        return super()._missing_(value)


class PurposeLineage(Enum):
    BASIC = "basic"
    WINDOW = "window"
    AGGREGATE = "aggregate"
    FILTER = "filter"
    CONSTANT = "constant"
    UNNEST = "unnest"
    ROOT = "root"
    ROWSET = "rowset"
    MULTISELECT = "multiselect"
    MERGE = "merge"


class Granularity(Enum):
    SINGLE_ROW = "single_row"
    MULTI_ROW = "multi_row"


class Modifier(Enum):
    PARTIAL = "Partial"
    OPTIONAL = "Optional"
    HIDDEN = "Hidden"
    NULLABLE = "Nullable"

    @classmethod
    def _missing_(cls, value):
        strval = str(value)
        if strval == "~":
            return Modifier.PARTIAL
        return super()._missing_(value=strval.capitalize())


class JoinType(Enum):
    INNER = "inner"
    LEFT_OUTER = "left outer"
    FULL = "full"
    RIGHT_OUTER = "right outer"
    CROSS = "cross"


class Ordering(Enum):
    ASCENDING = "asc"
    DESCENDING = "desc"


class WindowType(Enum):
    ROW_NUMBER = "row_number"
    RANK = "rank"
    LAG = "lag"
    LEAD = "lead"
    SUM = "sum"
    MAX = "max"
    MIN = "min"
    AVG = "avg"
    COUNT = "count"
    COUNT_DISTINCT = "count_distinct"


class WindowOrder(Enum):
    ASCENDING = "top"
    DESCENDING = "bottom"


class FunctionType(Enum):
    # custom
    CUSTOM = "custom"

    # structural
    UNNEST = "unnest"
    ALIAS = "alias"

    # Generic
    CASE = "case"
    CAST = "cast"
    CONCAT = "concat"
    CONSTANT = "constant"
    COALESCE = "coalesce"
    IS_NULL = "isnull"

    # COMPLEX
    INDEX_ACCESS = "index_access"
    ATTR_ACCESS = "attr_access"

    # TEXT AND MAYBE MORE
    SPLIT = "split"

    # Math
    DIVIDE = "divide"
    MULTIPLY = "multiply"
    ADD = "add"
    SUBTRACT = "subtract"
    MOD = "mod"
    ROUND = "round"
    ABS = "abs"

    # Aggregates
    ## group is not a real aggregate - it just means group by this + some other set of fields
    ## but is here as syntax is identical
    GROUP = "group"

    COUNT = "count"
    COUNT_DISTINCT = "count_distinct"
    SUM = "sum"
    MAX = "max"
    MIN = "min"
    AVG = "avg"
    LENGTH = "len"

    # String
    LIKE = "like"
    ILIKE = "ilike"
    LOWER = "lower"
    UPPER = "upper"
    SUBSTRING = "substring"
    STRPOS = "strpos"

    # Dates
    DATE = "date"
    DATETIME = "datetime"
    TIMESTAMP = "timestamp"

    # time
    SECOND = "second"
    MINUTE = "minute"
    HOUR = "hour"
    DAY = "day"
    DAY_OF_WEEK = "day_of_week"
    WEEK = "week"
    MONTH = "month"
    QUARTER = "quarter"
    YEAR = "year"

    DATE_PART = "date_part"
    DATE_TRUNCATE = "date_truncate"
    DATE_ADD = "date_add"
    DATE_DIFF = "date_diff"

    # UNIX
    UNIX_TO_TIMESTAMP = "unix_to_timestamp"

    # CONSTANTS
    CURRENT_DATE = "current_date"
    CURRENT_DATETIME = "current_datetime"


class FunctionClass(Enum):
    AGGREGATE_FUNCTIONS = [
        FunctionType.MAX,
        FunctionType.MIN,
        FunctionType.SUM,
        FunctionType.AVG,
        FunctionType.COUNT,
        FunctionType.COUNT_DISTINCT,
    ]
    SINGLE_ROW = [
        FunctionType.CONSTANT,
        FunctionType.CURRENT_DATE,
        FunctionType.CURRENT_DATETIME,
    ]


class Boolean(Enum):
    TRUE = "true"
    FALSE = "false"


class BooleanOperator(Enum):
    AND = "and"
    OR = "or"


class ComparisonOperator(Enum):
    LT = "<"
    GT = ">"
    EQ = "="
    IS = "is"
    IS_NOT = "is not"
    GTE = ">="
    LTE = "<="
    NE = "!="
    IN = "in"
    NOT_IN = "not in"
    # TODO: deprecate for contains?
    LIKE = "like"
    ILIKE = "ilike"
    CONTAINS = "contains"
    ELSE = "else"

    @classmethod
    def _missing_(cls, value):
        if not isinstance(value, list) and " " in str(value):
            value = str(value).split()
        if isinstance(value, list):
            processed = [str(v).lower() for v in value]
            if processed == ["not", "in"]:
                return ComparisonOperator.NOT_IN
            if processed == ["is", "not"]:
                return ComparisonOperator.IS_NOT
            if value == ["in"]:
                return ComparisonOperator.IN
        return super()._missing_(str(value).lower())


class DatePart(Enum):
    MONTH = "month"
    YEAR = "year"
    WEEK = "week"
    DAY = "day"
    QUARTER = "quarter"
    HOUR = "hour"
    MINUTE = "minute"
    SECOND = "second"

    @classmethod
    def _missing_(cls, value):
        if isinstance(value, str) and value.lower() != value:
            return DatePart(value.lower())
        return super()._missing_(value)


class SourceType(Enum):
    FILTER = "filter"
    SELECT = "select"
    MERGE = "merge"
    ABSTRACT = "abstract"
    DIRECT_SELECT = "direct_select"
    GROUP = "group"
    WINDOW = "window"
    UNNEST = "unnest"
    CONSTANT = "constant"
    ROWSET = "rowset"


class ShowCategory(Enum):
    MODELS = "models"
    CONCEPTS = "concepts"
